const child_process = require("child_process");
const fs = require("fs");
const path = require("path");
const Logger = require("./logger");
const MaxUtils = require("./utils");

const { ERROR_CODES, NPM_CMD_PATH, PATH_ENV_SEPARATOR } = require("./constants");

const STATES = Object.freeze({
	IDLE: 0,
	RUNNING: 1,
	CLOSED: 2,
	ERROR: -1
});

// We only allow one command per Node Project at a time
const lockedFolders = new Set();
class NSNPMCommand {

	constructor(options = {}) {

		if (!options.maxEnv) {
			const err = new Error("Invalid max environment for running npm command");
			err.code = ERROR_CODES.INVALID_MAX_ENV;
			err.details = JSON.stringify(options);
			throw err;
		}

		this._args = options.args || [];

		// Avoid interactive CLI for creating package.json as it's such a common task to init the package.json file from
		// within Max
		if (this._args.length && this._args[0] === "init" && !this._args.includes("--yes") && !this._args.includes("-y")) {
			this._args.push("--yes");
		}

		// Make sure to make npm use the correct node binary
		this._args.push("--scripts-prepend-node-path=true");

		this._cwd = options.path;
		this._id = MaxUtils.uniqID();

		this._executable = options.executable;
		this._npmPath = options.npmPath;

		this._err = null;
		this._state = STATES.IDLE;
		this._maxEnv = options.maxEnv;
	}

	get _npmDebugPath() {
		return path.join(this._cwd, `npm-debug-${this._id}.log`);
	}

	get maxEnv() {
		return this._maxEnv;
	}

	_unlockPath() {
		lockedFolders.delete(this.cwd);
		return true;
	}

	_lockPath() {
		if (lockedFolders.has(this.cwd)) return false;
		lockedFolders.add(this.cwd);
		return true;
	}

	_createNPMWriteStream() {
		return new Promise((resolve, reject) => {
			const stream = fs.createWriteStream(this._npmDebugPath);

			stream.once("error", err => {
				stream.removeAllListeners();
				Logger.debug("Unable to open WriteStream for NPM command with fd at %s. Skipping log output...", this._npmDebugPath);
				return void resolve(null);
			});

			stream.once("open", () => {
				stream.removeAllListeners("error");
				return void resolve(stream);
			});
		});
	}

	_removeNPMDebug() {
		return new Promise((resolve, reject) => {
			fs.access(this._npmDebugPath, (err) => {
				if (err) {
					// assume the file doesn't exists
					resolve();
					return;
				}

				// remove the file
				fs.unlink(this._npmDebugPath, (unlinkErr) => {
					if (unlinkErr) {
						const resErr = new Error(`Can't remove npm-debug-log at ${this.cwd}`);
						resErr.code = ERROR_CODES.NPM_DEBUG_NO_ACCESS;
						resErr.details = unlinkErr.message;
						reject(resErr);
						return;
					}

					resolve();
					return;
				});
			});
		});
	}

	_runNPMCommand(stderr = null) {
		return new Promise((resolve, reject) => {
			const proc = child_process.spawn(
				this.npmExecutablePath,
				this._args,
				{
					cwd: this.cwd,
					detached: false,
					env: Object.assign({}, process.env, {
						MAX_ENV: this.maxEnv,
						PATH: [
							path.dirname(this.npmExecutablePath),
							path.dirname(this.nodeExecutablePath),
							process.env.PATH
						].join(PATH_ENV_SEPARATOR)
					}),
					// shell: true,
					stdio: ["ignore", "ignore", stderr || "ignore"]
				}
			);

			proc.once("error", (err) => {
				proc.removeAllListeners();
				const errResult = new Error(`Error while attempting to run npm command in ${this.cwd}`);
				errResult.code = ERROR_CODES.NPM_EXEC_ERROR;
				errResult.details = err.message;
				return reject(errResult);
			});

			proc.once("close", (code) => {
				proc.removeAllListeners();

				if (code !== 0) {
					const err = new Error(`NPM exited with  non-zero code "${code}" at ${this.cwd}`);
					err.code = ERROR_CODES.NPM_EXEC_ERROR;
					err.log = stderr ? this._npmDebugPath : "";
					return reject(err);
				}

				return resolve();
			});
		});
	}

	get id() {
		return this._id;
	}

	get nodeExecutablePath() {
		return this._executable ? this._executable.trim() : process.execPath;
	}

	get npmExecutablePath() {
		return this._npmPath ? this._npmPath.trim() : NPM_CMD_PATH;
	}

	get status() {
		return this._state;
	}

	get cwd() {
		return this._cwd;
	}

	destroy() {
		// No-Op
	}

	async perform() {

		if (!this._lockPath()) {
			const err = new Error(`Can't perform simultaneous NPM command for folder ${this.cwd}`);
			err.code = ERROR_CODES.NO_SIMULTANEOUS_NPM;
			throw err;
		}

		// Validate Script Path
		if (!fs.existsSync(this.cwd)) {

			const err = new Error(`Failed to start process. Folder ${this.cwd} not found.`);
			err.code = ERROR_CODES.FILE_NOT_FOUND;
			this._err = err;
			this._state = err;
			this._unlockPath();

			throw err;
		}

		// Validate Executable Paths
		if (!fs.existsSync(this.nodeExecutablePath)) {

			const err = new Error(`Failed to start process. Node binary ${this.nodeExecutablePath} not found.`);
			err.code = ERROR_CODES.INVALID_NODE_BINARY;
			this._err = err;
			this._state = err;
			this._unlockPath();

			throw err;
		}

		if (!fs.existsSync(this.npmExecutablePath)) {

			const err = new Error(`Failed to start process. NPM binary ${this.npmExecutablePath} not found.`);
			err.code = ERROR_CODES.INVALID_NPM_BINARY;
			this._err = err;
			this._state = err;

			this._unlockPath();
			throw err;
		}

		this._state = STATES.RUNNING;

		try {
			const debugStream = await this._createNPMWriteStream();

			debugStream.once("error", (err) => {
				Logger.debug("WriteStream Error for NPM command with fd at %s", this._npmDebugPath);
			});

			await this._runNPMCommand(debugStream);
			this._state = STATES.CLOSED;
			await this._removeNPMDebug();
			return;

		} catch (err) {
			this._state = STATES.ERROR;
			this._err = err;

			throw err;
		} finally {
			this._unlockPath();
		}
	}
}

NSNPMCommand.STATES = STATES;
module.exports = NSNPMCommand;
